#include <stdlib.h>
#include <time.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <limits.h>

#include "data.h"
#include "config.h"
#include "util.h"
#include "xmlparser.h"
#include "textdatafile.h"
#include "xmlout.h"

#define COUNTERS 6
static char *counters_str[] = { "\0", "today", "week", "month", "year", "total" };
enum counters { now, total, today, week, month, year };
int SortByColumn = 1;

struct ip_entry
{
  unsigned ip;
  unsigned long long upload[COUNTERS];
  unsigned long long download[COUNTERS];
  unsigned char flags;
  struct ip_entry *next;
};

TrafficData::TrafficData()
{
  TimeStamp = 0;
  First = NULL;
  Last = NULL;
  Entries = 0;
}

TrafficData::~TrafficData()
{
  freeMemory();
}

int TrafficData::counterNumber(const char *counter)
{
  int i;
  for (i = 0; i < COUNTERS; i++)
  {
    if (strcasecmp(counter, counters_str[i]) == 0)
      return i;
  }

  return -1;
}

bool TrafficData::setTimeStamp(const char *time)
{
  if (!time)
    return false;

  TimeStamp = strtoul(time, NULL, 10);

  return true;
}

bool TrafficData::addIP(unsigned int ip)
{
  struct ip_entry *new_ip = new ip_entry;
  if (new_ip)
  {
    int i;

    new_ip->ip = ip;

    for (i = 0; i < COUNTERS; i++)
    {
      new_ip->upload[i] = 0;
      new_ip->download[i] = 0;
    }

    new_ip->next = NULL;
    new_ip->flags = 0;

    if (!First)
      First = new_ip;
    if (Last)
      Last->next = new_ip;
    Last = new_ip;
    Entries++;

    return true;
  }
  return false;
}

bool TrafficData::addFlags(unsigned int ip, unsigned flag)
{
  struct ip_entry *entry = findIP(ip);

  if (!entry)
  {
    if (!addIP(ip))
      return false;
    entry = Last;
  }

  entry->flags |= flag;
  return true;
}

int TrafficData::getFlags(unsigned int ip)
{
  struct ip_entry *entry = findIP(ip);

  if (!entry)
    return -1;

  return entry->flags;
}

struct ip_entry *TrafficData::findIP(unsigned int ip)
{
  struct ip_entry *entry = First;
  while (entry)
  {
    if (entry->ip == ip)
      break;
    entry = entry->next;
  }
  return entry;
}

bool TrafficData::setUpload(unsigned int ip, int counter,
                           unsigned long long bytes)
{
  struct ip_entry *entry = findIP(ip);
  if (!entry)
  {
    if (!addIP(ip))
      return false;
    entry = Last;
  }

  entry->upload[counter] = bytes;
  return true;
}

bool TrafficData::setUpload(unsigned int ip, const char *counter,
                           const char *byte_str)
{
  if (!byte_str || !*byte_str)
    return false;

  int which = counterNumber(counter);
  unsigned long long bytes = strtoull(byte_str, NULL, 10);

  if (bytes == ULONG_LONG_MAX)
    return false;
  if (which < 0)
    return false;

  return setUpload(ip, which, bytes);
}

bool TrafficData::setDownload(unsigned int ip, int counter,
                             unsigned long long bytes)
{
  struct ip_entry *entry = findIP(ip);
  if (!entry)
  {
    if (!addIP(ip))
      return false;
    entry = Last;
  }

  entry->download[counter] = bytes;
  return true;
}

bool TrafficData::setDownload(unsigned int ip, const char *counter,
                             const char *byte_str)
{
  if (!byte_str || !*byte_str)
    return false;

  int which = counterNumber(counter);
  unsigned long long bytes = strtoull(byte_str, NULL, 10);

  if (bytes == ULONG_LONG_MAX)
    return false;
  if (which < 0)
    return false;

  return setDownload(ip, which, bytes);
}

void TrafficData::freeMemory()
{
  struct ip_entry *entry = First;
  struct ip_entry *old;

  while (entry)
  {
    old = entry;
    entry = entry->next;
    delete old;
  }
}

int compareEntries(const void *x, const void *y)
{
  struct ip_entry *a = *(struct ip_entry **) x;
  struct ip_entry *b = *(struct ip_entry **) y;

  if (SortByColumn == 0)
  {
/*              printf("%s ", string_ip(a->ip));
                printf("%s ", string_ip(b->ip));*/
    if (a->ip < b->ip)
      return -1;
    else if (a->ip == b->ip)
      return 0;
    else if (a->ip > b->ip)
      return 1;
  }
  else
  {
    int c = SortByColumn - 1;
    int i = (c / 2) + 1;
    if (c & 0x1)
    {
      if (a->download[i] < b->download[i])
        return 1;
      else if (a->download[i] == b->download[i])
        return 0;
      else //if (a->download[i] > b->download[i])
        return -1;
    }
    else
    {
      if (a->upload[i] < b->upload[i])
        return 1;
      else if (a->upload[i] == b->upload[i])
        return 0;
      else //if (a->upload[i] > b->upload[i])
        return -1;
    }
  }
  return 0; 
}

TrafficTableIterator *TrafficData::Sort(int column)
{
  if (!Entries)
    return NULL;

  struct ip_entry **table =
    (struct ip_entry **) malloc(sizeof(struct ip_entry *) * Entries);
  if (!table)
  {
    fprintf(stderr, "Out of memory, can't sort.\n");
    return NULL;
  }

  SortByColumn = column;
  int i = 0;
  struct ip_entry *entry = First;
  while (entry)
  {
    if (i >= Entries)
    {
      fprintf(stderr, "Incosistent database.\n");
      exit(-4);
    }

    table[i] = entry;
    entry = entry->next;
    i++;
  }

  qsort(table, Entries, sizeof(struct ip_entry *), compareEntries);

  return new TrafficTableIterator(table, Entries);
}

bool TrafficData::Read(int fileFormat, char* fileName)
{
  bool ret = false;
  switch(fileFormat)
  {
    case DATA_XML:
      {
        XMLParser* xmlp = new XMLParser();
        ret = xmlp->parseDataFile(this, fileName);
        delete xmlp;
      }
      break;
                                                                                
    case DATA_TEXT:
      TextDataFile *textData = new TextDataFile();
      ret = textData->Read(this, fileName);
      delete textData;
      break;
/*    case DATA_BINARY:
      BinaryDataFile *binaryData = new BinaryDataFile();
      ret = binaryData->Read(data, fileName);
      break;
                                                                                
    case DATA_RRD:
      //do nothing
      break;
      */
  };
                                                                                
  return ret;
}

bool TrafficData::Save(int fileFormat, char* fileName)
{
  bool ret=false;
  switch(fileFormat)
  {
    case DATA_XML:
      {
        XMLOut *xmlout = new XMLOut();
        ret = xmlout->Save(fileName, getIterator());
        delete xmlout;
      }
      break;
      
    case DATA_TEXT:
      {
	TextDataFile *txt = new TextDataFile();
	ret = txt->Save(this, fileName);
	delete txt;
      }
      break;
    case DATA_BINARY:
    case DATA_RRD:
      //do nothing
      break;
  } 

  return ret;
}

void TrafficData::Print()
{
  struct ip_entry *entry = First;
  int i;

  while (entry)
  {
    printf("IP:%s FL: %u IN:", string_ip(entry->ip), entry->flags);

    for (i = 0; i < COUNTERS; i++)
      printf(" %llu", entry->upload[i]);

    printf(" OUT:");
    for (i = 0; i < COUNTERS; i++)
      printf(" %llu", entry->download[i]);

    printf("\n");

    entry = entry->next;
  }
}

int TrafficData::getCounterNumber()
{
  return COUNTERS;
}

int TrafficData::RemoveUnlisted(void(*func) (unsigned, unsigned))
{
  ip_entry* c = First;
  ip_entry *p = NULL, *n;
  unsigned i=0;

  c = First; 
  while(c)
  {
    if(!(c->flags & (F_ARP | F_LBCFREE | F_LIST)))
    { //remove unlisted entry
      n = c->next;

      (func)(c->ip, c->flags);
      
      if(p)
	p->next = n;
      else
	First = n;

      if(!n)
	Last = p;
	
      delete c;

      c = n;
      i++;
    }
    else
    { 
      p = c;
      c = c->next;
    }
  }

  return i;
}

TrafficDataIterator *TrafficData::getIterator()
{
  if (!First)
    return NULL;
  return new TrafficDataIterator(First);
}


/* Data Entries (unsorted) Iterator */
TrafficDataIterator::TrafficDataIterator(struct ip_entry * first)
{
  Current = first;
}

int TrafficDataIterator::Next()
{
  if (!Current->next)
    return 0;

  Current = Current->next;

  return 1;
}

/*int TrafficDataIterator::Previous()
{
        if(!Current->previous) return 0;
                                                                                                                               
        Current = Current->previous;
                                                                                                                               
        return 1;
}*/

unsigned TrafficDataIterator::getIP()
{
  return Current->ip;
}

unsigned TrafficDataIterator::getFlags()
{
  return Current->flags;
}

unsigned long long TrafficDataIterator::getUpload(int which)
{
  return Current->upload[which];
}

unsigned long long TrafficDataIterator::getDownload(int which)
{
  return Current->download[which];
}

unsigned TrafficDataIterator::getCounterNumber()
{
  return COUNTERS;
}

char *TrafficDataIterator::getCounterName(int which)
{
  return counters_str[which];
}

/* Table (sorted) iterator */

TrafficTableIterator::TrafficTableIterator(struct ip_entry ** table,
                                           int entries)
{
  Table = table;
  Entries = entries;
  Current = 0;
}

TrafficTableIterator::~TrafficTableIterator()
{
  free(Table);
}

int TrafficTableIterator::Next()
{
  Current++;
  if (Current >= Entries)
    return 0;

  return 1;
}

unsigned int TrafficTableIterator::getIP()
{
  return Table[Current]->ip;
}

unsigned TrafficTableIterator::getFlags()
{
  return Table[Current]->flags;
}

unsigned long long TrafficTableIterator::getUpload(int which)
{
  return Table[Current]->upload[which];
}

unsigned long long TrafficTableIterator::getDownload(int which)
{
  return Table[Current]->download[which];
}

unsigned TrafficTableIterator::getCounterNumber()
{
  return COUNTERS;
}

char *TrafficTableIterator::getCounterName(int which)
{
  return counters_str[which];
}

unsigned TrafficTableIterator::getNumberOfEntries()
{
  return Entries;
}
