#include <stdio.h>
#include <string.h>
#include <error.h>
#include <errno.h>
#include <expat.h>

#ifndef XMLCALL
#if defined(_MSC_EXTENSIONS) && !defined(__BEOS__) && !defined(__CYGWIN__)
#define XMLCALL __cdecl
#elif defined(__GNUC__)
#define XMLCALL __attribute__((cdecl))
#else
#define XMLCALL
#endif
#endif

#include "config.h"
#include "data.h"
#include "xmlparser.h"
#include "util.h"

struct xml_element
{
  char *name;
  struct xml_element *previous;
  struct xml_element *next;
};

#define REQUIRED 0x1
#define LAST 0x2

struct param
{
  char *name;
  bool (Config::*func) (const char *);
  unsigned flags;
};

struct handler
{
  char *name;
  param params[2];
};

param iptable_par[] = { "location", &Config::setIpTablesLocation, REQUIRED | LAST};

handler ConfigHandlers[] = {
  {"iptables", { "location", &Config::setIpTablesLocation, REQUIRED | LAST}},
  {"rrdtool", { "location", &Config::setRrdToolLocation, REQUIRED | LAST}},
  {"datadir", {"location", &Config::setDataDirLocation, REQUIRED | LAST}},
  {"input", { { "type", &Config::addInputType, REQUIRED },
              { "name", &Config::addInputName, REQUIRED | LAST }}},
  {"arp-device", { "add", &Config::addDevice, REQUIRED | LAST}},
  {"datafile", { {"format", &Config::setDataFormat, REQUIRED } ,
		 { "name", &Config::setDataFileName, LAST}}},
  {"olddatafile",{ {"format", &Config::setOldDataFormat, REQUIRED },
		{ "name", &Config::setOldDataFileName, LAST}}},
  {"htmltemplate", { "location", &Config::setHTMLTemplate, REQUIRED | LAST}},
  {"wwwrefresh", { "time", &Config::setWWWRefresh, REQUIRED | LAST}},
  {"graphscript", { "url", &Config::setGraphScriptURL, REQUIRED | LAST}},
  {"options", { { "remove_unlisted_addresses", &Config::setRemoveUnlistedAddresses, 0 },
		{ "delete_rrd", &Config::setDeleteRRDs, LAST }}},
//  {"datanew", { "format", &Config::setDataFormat, REQUIRED | LAST}},
//  {"dataold", { "format", &Config::setOldDataFormat, REQUIRED | LAST}},
  {0}
};

void XMLCALL start(void *data, const char *el, const char **attr)
{
  XMLParser *xp = (XMLParser *) data;
  xp->Start(el, attr);
}

void XMLCALL end(void *data, const char *el)
{
  XMLParser *xp = (XMLParser *) data;
  xp->End(el);
}

XMLParser::XMLParser()
{
  Root = NULL;
  IP = 0;
}

const char *XMLParser::getParam(char *name, const char **attr)
{
  int i;
  for (i = 0; attr[i]; i += 2)
  {
    if (strcasecmp(attr[i], name) == 0)
      return (attr[i + 1]);
  }

  return NULL;
}

void XMLParser::HandleConfigElement(const char *el, const char **attr)
{
  int i = 0;
  while (ConfigHandlers[i].name)
  {
    if (strcasecmp(ConfigHandlers[i].name, el) == 0)
    {
      for(int j=0;;j++)
      {
        const char *par = getParam(ConfigHandlers[i].params[j].name, attr);
        if (!par)
	{
	  if(ConfigHandlers[i].params[j].flags & REQUIRED)
          {
            fprintf(stderr,
                   "Error element %s is missing required attribute %s.\n", el,
                    ConfigHandlers[i].params[j].name);
	    exit(-1);
          }
	}
	else
	{
          int result = (Conf->*ConfigHandlers[i].params[j].func) (par);
          if (!result)
          {
            fprintf(stderr,
                    "Error while processing element %s attribute %s value '%s'\n",
                    el, ConfigHandlers[i].params[j].name, par);
	    exit(-1);
          }
	}

        if(ConfigHandlers[i].params[j].flags & LAST) 
	  break;
      }
      return;
    }
    i++;
  }
  fprintf(stderr, "Error unknown config element %s.\n", el);
//  exit(-1);
}

void XMLParser::Start(const char *el, const char **attr)
{
  xml_element *newel = new xml_element;

  newel->previous = Current;
  newel->next = NULL;
  newel->name = strdup(el);

  if (Current)
    Current->next = newel;
  else
    Root = newel;

  Current = newel;

  /*printf("%s", el);
     for (i = 0; attr[i]; i += 2) {
     printf(" %s='%s'", attr[i], attr[i + 1]);
     }
     printf("\n"); */

  if (Root && Root->next)
  {
    if (Conf && strcasecmp(Root->name, "config") == 0)
    {
      if (Root->next != Current)
      {
        fprintf(stderr, "Parse error at line %d\n",
                XML_GetCurrentLineNumber(Parser));
//      XML_StopParser(Parser, false);
	exit(-1);
      }

      HandleConfigElement(el, attr);
    }

    if (Data && strcasecmp(Root->name, "data") == 0)
    {
      if (Root->next == NULL || Root->next == Current)
      {
        if (strcasecmp(Current->name, "ip") == 0)
        {
          IP = parse_ip((char *) getParam("addr", attr));
//          Data->addIP(IP);        
        }
      }
      else
      {
        if (strcasecmp(Current->name, "rx") == 0
            || strcasecmp(Current->name, "tx") == 0)
        {
          //ok
        }
        else if (strcasecmp(Current->previous->name, "upload") == 0
                 || strcasecmp(Current->previous->name, "rx") == 0)
        {
          if (!Data->setUpload(IP, Current->name, getParam("bytes", attr)))
            fprintf(stderr, "Error unknown upload element %s.\n", el);
	    exit(-1);

        }
        else if (strcasecmp(Current->previous->name, "download") == 0
                 || strcasecmp(Current->previous->name, "tx") == 0)
        {
          if (!Data->setDownload(IP, Current->name, getParam("bytes", attr)))
            fprintf(stderr, "Error unknown download element %s.\n", el);
	    exit(-1);
        }
      }
    }
  }
  else
  {
    if (Data && strcasecmp(Current->name, "data") == 0)
    {
      Data->setTimeStamp(getParam("timestamp", attr));
    }

    Root = Current;
  }
}                               /* End of start handler */

void XMLParser::End(const char *el)
{
//  printf("end: freeing %s\n", current->name);
  xml_element *prevel = Current->previous;
  free(Current->name);
  delete Current;

  Current = prevel;
  if (Current)
    Current->next = NULL;
  else
    Root = NULL;

}                               /* End of end handler */

bool XMLParser::parseConfigFile(Config* conf, char* configFile)
{
  bool ret;

  Conf = conf;
  ret = parseFile(configFile);
  Conf = NULL;

  return ret;
}

bool XMLParser::parseDataFile(TrafficData* data, char* configFile)
{
  bool ret;
                                                                                
  Data = data;
  ret = parseFile(configFile);
  Data = NULL;
                                                                                
  return ret;
}


bool XMLParser::parseFile(char *file)
{
  FILE *f = fopen(file, "rt");
  if (!f)
  {
    fprintf(stderr, "Nelze otevrit %s: %s\n", file, strerror(errno));
    return false;
  }

  Parser = XML_ParserCreate(NULL);
  if (!Parser)
  {
    fprintf(stderr, "Couldn't allocate memory for parser\n");
    exit(-1);
  }

  XML_SetElementHandler(Parser, start, end);
  XML_SetUserData(Parser, this);
  XML_SetParamEntityParsing(Parser,
                            XML_PARAM_ENTITY_PARSING_UNLESS_STANDALONE);

  char *s = NULL;
  size_t len = 0;
  int done;

  while (getline(&s, &len, f) != -1)
  {
    done = feof(f);
    if (XML_Parse(Parser, s, strlen(s), done) == XML_STATUS_ERROR)
    {
      fprintf(stderr, "Parse error at line %d:\n%s\n",
              XML_GetCurrentLineNumber(Parser),
              XML_ErrorString(XML_GetErrorCode(Parser)));
      fprintf(stderr, "\'%s\'", s);
      exit(-1);
    }
  }

  *s = 0;
  XML_ParserFree(Parser);
  if (s)
    free(s);
  fclose(f);

  return true;
}
